"""
A short script to list names of distributions modified since some
typeshed commit.

This is used by scheduled GitHub auto-upload action to only upload
modified stub packages.
"""

import argparse
import os
import subprocess


def main(typeshed_dir: str, commit: str) -> list[str]:
    """List all distributions that changed since commit."""
    assert typeshed_dir.endswith(os.sep + "typeshed")
    git = subprocess.run(
        ["git", "diff", "--no-renames", "--name-only", "HEAD", commit],
        capture_output=True,
        text=True,
        cwd=typeshed_dir,
        check=True,
    )
    changed = set()
    for file in git.stdout.splitlines():
        # Third party stubs live in typeshed/stubs/...
        if file.startswith("stubs" + os.path.sep):
            _, distribution, *_ = file.split(os.path.sep)
            changed.add(distribution)
    return sorted(changed)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    parser.add_argument(
        "previous_commit", help="Previous typeshed commit for which we performed upload"
    )
    args = parser.parse_args()
    for distribution in main(args.typeshed_dir, args.previous_commit):
        print(distribution)



from __future__ import annotations

import functools
import graphlib
import os
import re
import tarfile
import tempfile
import tomllib
import urllib.parse
from collections.abc import Generator, Iterable
from glob import glob
from pathlib import Path
from typing import Any, Optional

import requests
from packaging.requirements import Requirement
from packaging.specifiers import InvalidSpecifier, Specifier

from .const import META, THIRD_PARTY_NAMESPACE, TYPES_PREFIX, UPLOADED_PATH


class InvalidRequires(Exception):
    pass


class Metadata:
    def __init__(self, distribution: str, data: dict[str, Any]):
        assert not distribution.startswith(TYPES_PREFIX)
        self._alleged_upstream_distribution = distribution
        self.data = data

    @property
    def upstream_distribution(self) -> Optional[str]:
        # TODO: add a field to METADATA.toml if the stubs are for a package
        # that does not exist on PyPI
        if self._alleged_upstream_distribution == "gdb":
            return None
        return self._alleged_upstream_distribution

    @property
    def stub_distribution(self) -> str:
        distribution = self.data.get(
            "stub_distribution", TYPES_PREFIX + self._alleged_upstream_distribution
        )
        assert isinstance(distribution, str)
        return distribution

    @property
    def version_spec(self) -> Specifier:
        # The "version" field in METADATA.toml isn't actually a version, it's more
        # like a specifier, e.g. we allow it to contain wildcards.
        version = self.data["version"]
        assert isinstance(version, str)
        if version[0].isdigit():
            version = f"=={version}"
        spec = Specifier(version)
        assert spec.operator in {"==", "~="}
        return spec

    @property
    def _unvalidated_requires(self) -> list[Requirement]:
        return [Requirement(req) for req in self.data.get("requires", [])]

    @property
    def _unvalidated_requires_typeshed(self) -> list[Requirement]:
        typeshed = uploaded_packages.read()
        return [
            r for r in self._unvalidated_requires if canonical_name(r.name) in typeshed
        ]

    @functools.cached_property
    def requires_typeshed(self) -> list[Requirement]:
        reqs = self._unvalidated_requires_typeshed
        for req in reqs:
            verify_typeshed_req(req)
        return reqs

    @property
    def _unvalidated_requires_external(self) -> list[Requirement]:
        typeshed = uploaded_packages.read()
        return [
            r
            for r in self._unvalidated_requires
            if canonical_name(r.name) not in typeshed
        ]

    @functools.cached_property
    def requires_external(self) -> list[Requirement]:
        reqs = self._unvalidated_requires_external
        for req in reqs:
            verify_external_req(req, self.upstream_distribution)
        return reqs

    @property
    def extra_description(self) -> str:
        description = self.data.get("extra_description", "")
        assert isinstance(description, str)
        return description

    @property
    def obsolete_since(self) -> str | None:
        obsolete = self.data.get("obsolete_since")
        assert isinstance(obsolete, (str, type(None)))
        return obsolete

    @property
    def no_longer_updated(self) -> bool:
        updated = self.data.get("no_longer_updated", False)
        assert isinstance(updated, bool)
        return updated

    @property
    def upload(self) -> bool:
        upload = self.data.get("upload", True)
        assert isinstance(upload, bool)
        return upload

    @property
    def partial(self) -> bool:
        partial = self.data.get("partial_stub", False)
        assert isinstance(partial, bool)
        return partial

    @property
    def requires_python(self) -> str | None:
        req = self.data.get("requires_python", None)
        assert isinstance(req, (str, type(None)))
        verify_requires_python(req)
        return req

    @functools.cached_property
    def upstream_repository(self) -> str | None:
        ts_upstream_repo = self.data.get("upstream_repository")
        if not isinstance(ts_upstream_repo, str):
            # either typeshed doesn't list it for these stubs,
            # or it gives a non-str for the field (bad!)
            return None
        try:
            parsed_url = urllib.parse.urlsplit(ts_upstream_repo)
        except ValueError:
            return None
        if parsed_url.scheme != "https":
            return None
        return ts_upstream_repo


def read_metadata(typeshed_dir: str, distribution: str) -> Metadata:
    """Parse metadata from file."""
    assert not distribution.startswith(TYPES_PREFIX)
    path = os.path.join(typeshed_dir, THIRD_PARTY_NAMESPACE, distribution, META)
    with open(path, "rb") as f:
        data = tomllib.load(f)
    return Metadata(distribution=distribution, data=data)


def canonical_name(name: str) -> str:
    # https://peps.python.org/pep-0503/#normalized-names
    return re.sub(r"[-_.]+", "-", name).lower()


class _UploadedPackages:
    def __init__(self, file_path: str) -> None:
        self._file_path = file_path
        self._cached: Optional[list[str]] = None

    def read(self) -> set[str]:
        if self._cached is not None:
            return set(map(canonical_name, self._cached))
        with open(self._file_path) as f:
            self._cached = f.read().splitlines()
        return set(map(canonical_name, self._cached))

    def add(self, distribution: str) -> None:
        assert not distribution.startswith(TYPES_PREFIX)
        stub_dist = TYPES_PREFIX + distribution
        if canonical_name(stub_dist) not in self.read():
            with open(self._file_path) as f:
                current = f.read().splitlines()
            current.append(stub_dist)
            current.sort()
            with open(self._file_path, "w") as f:
                f.write("\n".join(current))
            self._cached = None


uploaded_packages = _UploadedPackages(UPLOADED_PATH)


def strip_types_prefix(dependency: str) -> str:
    if not dependency.startswith(TYPES_PREFIX):
        raise ValueError("Expected dependency on a typeshed package")
    return dependency.removeprefix(TYPES_PREFIX)


def verify_typeshed_req(req: Requirement) -> None:
    if not req.name.startswith(TYPES_PREFIX):
        raise InvalidRequires(f"Expected dependency {req} to start with {TYPES_PREFIX}")

    if canonical_name(req.name) not in uploaded_packages.read():
        raise InvalidRequires(
            f"Expected dependency {req} to be uploaded from stub_uploader"
        )

    # TODO: make sure that if a typeshed distribution depends on other typeshed stubs,
    # the upstream depends on the upstreams corresponding to those stubs.
    # See https://github.com/typeshed-internal/stub_uploader/pull/61#discussion_r979327370


# Presence in the top 1000 PyPI packages could be a necessary but not sufficient criterion for
# inclusion in this allowlist.
# Note we could loosen our criteria once we address:
# https://github.com/typeshed-internal/stub_uploader/pull/61#discussion_r979327370
EXTERNAL_REQ_ALLOWLIST = {
    "Flask",
    "Flask-SQLAlchemy",
    "MarkupSafe",
    "Pillow",
    "Werkzeug",
    "arrow",
    "asgiref",
    "beautifulsoup4",
    "click",
    "cryptography",
    "django-stubs",
    "djangorestframework-stubs",
    "httpx",
    "matplotlib",
    "numpy",
    "pandas-stubs",
    "pygobject-stubs",
    "pyproj",
    "pytest",
    "referencing",
    "setuptools",
    "torch",
    "tree-sitter",
    "urllib3",
    "websockets",
    "wsproto",
}

# Map of external stub packages to their runtime equivalent.
# We check that the stubs actually depend on their runtime package.
EXTERNAL_RUNTIME_REQ_MAP = {
    "django-stubs": "django",
    "djangorestframework-stubs": "djangorestframework",
    "pandas-stubs": "pandas",
    "pygobject-stubs": "pygobject",
}


def validate_pypi_response(resp: requests.Response, req: Requirement) -> None:
    if resp.status_code != 200:
        raise InvalidRequires(
            f"Expected dependency {req} to be accessible on PyPI, but got {resp.status_code}"
        )


def temp_archive_path(sdist_data: dict[str, str], req: Requirement) -> str:
    tmpdir = tempfile.mkdtemp()
    archive_path = Path(tmpdir, sdist_data["filename"])

    resp = requests.get(sdist_data["url"], stream=True)
    validate_pypi_response(resp, req)
    with open(archive_path, "wb") as file:
        file.write(resp.raw.read())

    with tarfile.open(archive_path) as file_in:
        if hasattr(tarfile, "data_filter"):
            file_in.extraction_filter = tarfile.data_filter
        file_in.extractall(tmpdir)

    return tmpdir


def extract_sdist_requires(
    sdist_data: dict[str, str], req: Requirement
) -> Generator[Requirement, None, None]:
    tmpdir = temp_archive_path(sdist_data, req)

    # Only a single folder with "<package-version>.egg-info/requires.txt" in the archive should exist
    # but this supports possible edge-cases and doesn't require knowing any variable name in the path.
    requires_filepath = Path(tmpdir, "*", "*.egg-info", "requires.txt")
    matches = glob(str(requires_filepath))
    for match in matches:
        with open(match) as requires_file:
            lines = requires_file.readlines()
        for line in lines:
            # Skip empty lines and extras
            if line[0] not in {"\n", "["}:
                yield Requirement(line)


def extract_sdist_pyproject_requires(
    sdist_data: dict[str, str], req: Requirement
) -> Generator[Requirement, None, None]:
    tmpdir = temp_archive_path(sdist_data, req)

    # Find and process only the first pyproject.toml we come across:
    matches = list(Path(tmpdir).rglob("pyproject.toml"))
    if not matches:
        return
    pyproject = Path(matches[0])
    data = tomllib.loads(pyproject.read_text())

    # --- [dependency-groups] (PEP 735) ---
    groups = data.get("dependency-groups", {})
    if not isinstance(groups, dict):
        return
    for deps in groups.values():
        if not isinstance(deps, list):
            continue
        for dep in deps:
            if isinstance(dep, str):
                yield Requirement(dep)


def get_latest_sdist_data(pypi_data: dict[str, Any]) -> dict[str, Any] | None:
    """Return metadata for the latest sdist entry from PyPI API response."""
    return next(
        (
            url_data
            for url_data in reversed(pypi_data["urls"])
            if url_data["packagetype"] == "sdist"
        ),
        None,
    )


def verify_external_req(
    req: Requirement,
    upstream_distribution: Optional[str],
    *,
    _unsafe_ignore_allowlist: bool = False,  # used for tests
) -> None:
    """Verify that a non-typeshed dependency is valid.

    Raise InvalidRequires if the dependency is invalid.
    """

    verify_external_req_not_in_typeshed(req)
    verify_external_req_name(req)
    verify_external_req_in_allowlist(
        req, _unsafe_ignore_allowlist=_unsafe_ignore_allowlist
    )
    verify_external_req_stubs_require_its_runtime(req, upstream_distribution)


def verify_external_req_not_in_typeshed(req: Requirement) -> None:
    req_canonical_name = canonical_name(req.name)
    if req_canonical_name in uploaded_packages.read():
        raise InvalidRequires(
            f"Expected dependency {req} to not be uploaded from stub_uploader"
        )


def verify_external_req_name(req: Requirement) -> None:
    if req.name.startswith(TYPES_PREFIX):
        # technically this could be allowed, but it's very suspicious
        raise InvalidRequires(
            f"Expected dependency {req} to not start with {TYPES_PREFIX}"
        )


def verify_external_req_in_allowlist(
    req: Requirement,
    *,
    _unsafe_ignore_allowlist: bool = False,  # used for tests
) -> None:
    if req.name not in EXTERNAL_REQ_ALLOWLIST and not _unsafe_ignore_allowlist:
        msg = f"Expected dependency {req.name} to be present in the stub_uploader allowlist"
        if req.name in EXTERNAL_RUNTIME_REQ_MAP.values():
            maybe = next(
                k for k, v in EXTERNAL_RUNTIME_REQ_MAP.items() if v == req.name
            )
            msg += f". Did you mean {maybe}?"
        raise InvalidRequires(msg)


def verify_external_req_stubs_require_its_runtime(
    req: Requirement, upstream_distribution: str | None
) -> None:
    """Verify that an external stubs package requires its runtime package."""

    if upstream_distribution is None:
        raise InvalidRequires(
            f"There is no upstream distribution on PyPI, so cannot verify {req}"
        )

    resp = requests.get(f"https://pypi.org/pypi/{upstream_distribution}/json")
    validate_pypi_response(resp, req)
    data: dict[str, Any] = resp.json()

    # TODO: PyPI doesn't seem to have version specific requires_dist. This does mean we can be
    # broken by new releases of upstream packages, even if they do not match the version spec we
    # have for the upstream distribution.

    if not (
        req.name == upstream_distribution  # Allow `types-foo` to require `foo`
        or runtime_in_upstream_requires(req, data)
        or runtime_in_upstream_sdist_requires(req, data)
        or runtime_in_upstream_group_requires(req, data)
    ):
        runtime_req_name = EXTERNAL_RUNTIME_REQ_MAP.get(req.name, req.name)
        raise InvalidRequires(
            f"Expected dependency {runtime_req_name} to be listed in {upstream_distribution}'s "
            + "requires_dist, the sdist's *.egg-info/requires.txt or pyproject.toml"
        )


def runtime_in_upstream_requires(req: Requirement, data: dict[str, Any]) -> bool:
    """Return whether an external stubs package depends on its runtime package."""

    # TODO: PyPI doesn't seem to have version specific requires_dist. This does mean we can be
    # broken by new releases of upstream packages, even if they do not match the version spec we
    # have for the upstream distribution.

    runtime_req_name = EXTERNAL_RUNTIME_REQ_MAP.get(req.name, req.name)
    runtime_req_canonical_name = canonical_name(runtime_req_name)

    requires_dist = data["info"].get("requires_dist") or []
    return runtime_req_canonical_name in {
        canonical_name(Requirement(r).name) for r in requires_dist
    }


def runtime_in_upstream_sdist_requires(req: Requirement, data: dict[str, Any]) -> bool:
    """Return whether an external stubs package depends on its runtime package."""

    runtime_req_name = EXTERNAL_RUNTIME_REQ_MAP.get(req.name, req.name)
    runtime_req_canonical_name = canonical_name(runtime_req_name)

    sdist_data = get_latest_sdist_data(data)
    if sdist_data is None:
        return False

    return runtime_req_canonical_name in {
        canonical_name(r.name) for r in extract_sdist_requires(sdist_data, req)
    }


# Dependency groups (PEP 735) are not exposed via PyPI JSON or requires_dist.
# The only reliable way to detect them is by inspecting pyproject.toml in the sdist.
def runtime_in_upstream_group_requires(req: Requirement, data: dict[str, Any]) -> bool:
    """Return whether an external stubs package depends on its runtime package."""

    runtime_req_name = EXTERNAL_RUNTIME_REQ_MAP.get(req.name, req.name)
    runtime_req_canonical_name = canonical_name(runtime_req_name)

    sdist_data = get_latest_sdist_data(data)
    if sdist_data is None:
        return False

    return runtime_req_canonical_name in {
        canonical_name(r.name)
        for r in extract_sdist_pyproject_requires(sdist_data, req)
    }


def sort_by_dependency(
    typeshed_dir: str, distributions: Iterable[str]
) -> Generator[str, None, None]:
    # Just a simple topological sort. Unlike previous versions of the code, we do not rely
    # on this to perform validation, like requiring the graph to be complete.
    # We only use this to help with edge cases like multiple packages being uploaded
    # for the first time that depend on each other.
    ts: graphlib.TopologicalSorter[str] = graphlib.TopologicalSorter()

    dist_map: dict[str, str] = {}  # maps stub distribution name to directory name
    for dist in os.listdir(os.path.join(typeshed_dir, THIRD_PARTY_NAMESPACE)):
        metadata = read_metadata(typeshed_dir, dist)
        ts.add(
            metadata.stub_distribution,
            # Use _unvalidated_requires instead of requires_typeshed, in case we're uploading
            # a new package B that depends on another new package A. Sorting topologically means
            # that the A will be in uploaded_packages.txt by the time it comes to verify and
            # upload B.
            *[r.name for r in metadata._unvalidated_requires],
        )
        dist_map[metadata.stub_distribution] = dist

    # ts.static_order may contain external dependencies, so we filter them out
    ordered = [
        dist_map[stub_dist] for stub_dist in ts.static_order() if stub_dist in dist_map
    ]

    distributions = set(distributions)
    missing = distributions - set(ordered)
    assert not missing, f"Failed to find distributions {missing}"

    for dist in ordered:
        if dist in distributions:
            yield dist


def recursive_verify(metadata: Metadata, typeshed_dir: str) -> set[str]:
    # While metadata.requires_typeshed and metadata.requires_external will perform validation on the
    # stub distribution itself, it seems useful to be able to validate the transitive typeshed
    # dependency graph for a stub distribution
    _verified: set[str] = set()

    def _verify(metadata: Metadata) -> None:
        if metadata.stub_distribution in _verified:
            return
        _verified.add(metadata.stub_distribution)

        # calling these checks metadata's requires
        assert isinstance(metadata.requires_typeshed, list)
        assert isinstance(metadata.requires_external, list)

        # and recursively verify all our internal dependencies as well
        for req in metadata.requires_typeshed:
            _verify(read_metadata(typeshed_dir, strip_types_prefix(req.name)))

    _verify(metadata)
    return _verified


def verify_requires_python(requires_python: str | None) -> None:
    if requires_python is None:
        return
    try:
        specifier = Specifier(requires_python)
    except InvalidSpecifier as e:
        raise InvalidRequires(
            f"Invalid requires_python specifier: {requires_python}"
        ) from e
    if specifier.operator != ">=":
        raise InvalidRequires(
            f"Expected requires_python to be a '>=' specifier: {requires_python}"
        )



"""
Basic script to generate a wheel for a third-party distribution in typeshed.

This generates a stub package using METADATA.toml file for a given
distribution in typeshed stubs. Such package can be used by type checkers
to check code that uses the corresponding runtime Python package.

The generated wheel includes all type stubs (*.pyi files) and the METADATA.toml
itself, no other files can be included.

The types stubs live in https://github.com/python/typeshed/tree/main/stubs,
all fixes for types and metadata should be contributed there, see
https://github.com/python/typeshed/blob/main/CONTRIBUTING.md for more details.

This file also contains some helper functions related to wheel validation and upload.
"""

import argparse
import os
import os.path
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from textwrap import dedent

from stub_uploader.const import (
    CHANGELOG_PATH,
    META,
    TESTS_NAMESPACE,
    THIRD_PARTY_NAMESPACE,
)
from stub_uploader.metadata import Metadata, read_metadata
from stub_uploader.ts_data import TypeshedData, read_typeshed_data

CHANGELOG = "CHANGELOG.md"

SUFFIX = "-stubs"

PARTIAL_STUBS_DESCRIPTION = """
This stub package is marked as [partial](https://typing.python.org/en/latest/spec/distributing.html#partial-stub-packages).
If you find that annotations are missing, feel free to contribute and help complete them.
""".lstrip()

SETUP_TEMPLATE = dedent(
    """
from setuptools import setup

setup(package_data={package_data})
"""
)

PYPROJECT_TEMPLATE = dedent(
    """
[build-system]
build-backend = "setuptools.build_meta"
requires = ["setuptools>=77.0.3"]

[project]
name = "{stub_distribution}"
version = "{version}"
license = "Apache-2.0"
license-files = ["LICENSE"]
description = "Typing stubs for {distribution}"
readme = {{ text = \"\"\"\\\n{long_description}\n\"\"\", content-type = "text/markdown" }}
classifiers = [
  "Programming Language :: Python :: 3",
  "Typing :: Stubs Only",
]
requires-python = "{requires_python}"
dependencies = {requires}

[project.urls]
"Homepage" = "https://github.com/python/typeshed"
"GitHub" = "https://github.com/python/typeshed"
"Changes" = "https://github.com/typeshed-internal/stub_uploader/blob/main/data/changelogs/{distribution}.md"
"Issue tracker" = "https://github.com/python/typeshed/issues"
"Chat" = "https://gitter.im/python/typing"

[tool.setuptools]
packages = {packages}
include-package-data = false
"""
).lstrip()

NO_LONGER_UPDATED_TEMPLATE = """
*Note:* `{stub_distribution}` is unmaintained and won't be updated.
""".lstrip()

OBSOLETE_SINCE_TEXT_TEMPLATE = """
*Note:* The `{distribution}` package includes type annotations or type stubs
since version {obsolete_since}. Please uninstall the `{stub_distribution}`
package if you use this or a newer version.
""".lstrip()

DESCRIPTION_INTRO_TEMPLATE = """
## Typing stubs for {distribution}

This is a [type stub package](https://typing.python.org/en/latest/tutorials/external_libraries.html)
for the {formatted_distribution} package. It can be used by type checkers
to check code that uses `{distribution}`. This version of
`{stub_distribution}` aims to provide accurate annotations for
`{distribution}{typeshed_version_spec}`.
""".strip()

DESCRIPTION_OUTRO_TEMPLATE = """
This package is part of the [typeshed project](https://github.com/python/typeshed).
All fixes for types and metadata should be contributed there.
See [the README](https://github.com/python/typeshed/blob/main/README.md)
for more details. The source for this package can be found in the
[`stubs/{distribution}`](https://github.com/python/typeshed/tree/main/stubs/{distribution})
directory.

This package was tested with the following type checkers:
* [mypy](https://github.com/python/mypy/) {ts_data.mypy_version}
* [pyright](https://github.com/microsoft/pyright) {ts_data.pyright_version}

It was generated from typeshed commit
[`{commit}`](https://github.com/python/typeshed/commit/{commit}).
""".strip()


class BuildData:
    def __init__(self, typeshed_dir: str, distribution: str) -> None:
        self.distribution = distribution
        self.stub_dir = Path(typeshed_dir) / THIRD_PARTY_NAMESPACE / distribution


class PackageData:
    """Information about the packages of a distribution and their contents."""

    def __init__(self, base_path: Path, package_data: dict[str, list[str]]) -> None:
        self.base_path = base_path
        self.package_data = package_data

    def package_source_path(self, package_name: str) -> Path:
        """Return the relative path of a given package name in the typeshed directory.

        The package name can use dotted notation to address sub-packages.
        The top-level package name can optionally include the "-stubs" suffix.
        """
        top_level, *sub_packages = package_name.split(".")
        if top_level.endswith(SUFFIX):
            top_level = top_level[: -len(SUFFIX)]
        return Path(top_level, *sub_packages)

    def package_build_path(self, package_name: str) -> Path:
        """Return the relative path of a given package name in the build directory.

        The package name can use dotted notation to address sub-packages.
        The top-level package name will include the "-stubs" suffix.
        """
        top_level, *sub_packages = package_name.split(".")
        if not top_level.endswith(SUFFIX):
            top_level += SUFFIX
        return Path(top_level, *sub_packages)

    def is_single_file_package(self, package_name: str) -> bool:
        filename = package_name.split("-")[0] + ".pyi"
        return (self.base_path / filename).exists()

    @property
    def top_level_packages(self) -> list[str]:
        """Top level package names.

        These are the packages that are not sub-packages of any other package
        and includes namespace packages. Their name includes the "-stubs"
        suffix.
        """
        return list(self.package_data.keys())

    @property
    def top_level_non_namespace_packages(self) -> list[str]:
        """Top level non-namespace package names.

        This will return all packages that are not subpackages of any other
        package, other than namespace packages in dotted notation, e.g. if
        "flying" is a top level namespace package, and "circus" is a
        non-namespace sub-package, this will return ["flying-stubs.circus"].
        """
        packages: list[str] = []
        for top_level in self.top_level_packages:
            if self.is_single_file_package(top_level):
                packages.append(top_level)
            else:
                packages.extend(self._find_non_namespace_sub_packages(top_level))
        return packages

    def _find_non_namespace_sub_packages(self, package: str) -> list[str]:
        path = self.base_path / self.package_source_path(package)
        if is_namespace_package(path):
            sub_packages: list[str] = []
            for entry in path.iterdir():
                if entry.is_dir():
                    sub_name = package + "." + entry.name
                    sub_packages.extend(self._find_non_namespace_sub_packages(sub_name))
            return sub_packages
        else:
            return [package]

    def add_file(self, package: str, filename: str) -> None:
        """Add a file to a top-level package."""
        top_level, *sub_packages = package.split(".")
        assert top_level.endswith(SUFFIX)
        path = Path(*sub_packages, filename)
        self.package_data[top_level].append(str(path))


def is_namespace_package(path: Path) -> bool:
    return not (path / "__init__.pyi").exists()


def find_stub_files(top: str) -> list[str]:
    """Find all stub files for a given package, relative to package root.

    Raise if we find any unknown file extensions during collection.
    """
    result: list[str] = []
    for root, _, files in os.walk(top):
        for file in files:
            if file.endswith(".pyi"):
                name, _ = os.path.splitext(file)
                assert (
                    name.isidentifier()
                ), "All file names must be valid Python modules"
                result.append(os.path.relpath(os.path.join(root, file), top))
            elif file == "py.typed":
                result.append(os.path.relpath(os.path.join(root, file), top))
            elif not file.endswith((".md", ".rst")):
                # Allow having README docs, as some stubs have these (e.g. click).
                if (
                    subprocess.run(["git", "check-ignore", file], cwd=top).returncode
                    != 0
                ):
                    raise ValueError(f"Only stub files are allowed, not {file!r}")
    return sorted(result)


def copy_stubs(base_dir: Path, dst: Path) -> None:
    """Copy stubs for given distribution to the build directory.

    For packages change name by appending "-stubs" suffix (per the typing spec),
    also convert modules to trivial packages with a single __init__.pyi.
    """
    for entry in base_dir.iterdir():
        if entry.is_file():
            if entry.suffix != ".pyi":
                continue
            stub_dir = dst / (entry.stem + SUFFIX)
            stub_dir.mkdir(exist_ok=True)
            shutil.copy(entry, stub_dir / "__init__.pyi")
        else:
            if entry.name == TESTS_NAMESPACE:
                # Don't package tests for the stubs
                continue
            # Don't append the suffix if already present
            stub_dir = dst / ensure_suffix(entry.name, SUFFIX)
            shutil.copytree(entry, stub_dir, dirs_exist_ok=True)

        # We add original METADATA file in case some type-checking tool will want
        # to use it. Note that since it is not easy to package it outside of a package,
        # we copy it to every package in given distribution.
        shutil.copy(base_dir / META, stub_dir)


def ensure_suffix(s: str, suffix: str) -> str:
    """Ensure that a string ends with a suffix."""
    return s if s.endswith(suffix) else s + suffix


def create_py_typed(metadata: Metadata, pkg_data: PackageData, dst: Path) -> None:
    """Create py.typed files as necessary."""
    for package in pkg_data.top_level_non_namespace_packages:
        py_typed_path = dst / pkg_data.package_build_path(package) / "py.typed"
        py_typed_path.write_text("partial\n" if metadata.partial else "")


def copy_license(ts_data: TypeshedData, dst: Path) -> None:
    """Copy the license file from the typeshed repository to the build directory."""
    shutil.copy(ts_data.typeshed_path / "LICENSE", dst / "LICENSE")


def copy_changelog(distribution: str, dst: Path) -> None:
    """Copy changelog to the build directory."""
    try:
        shutil.copy(CHANGELOG_PATH / f"{distribution}.md", dst / CHANGELOG)
        with open(dst / "MANIFEST.in", "a") as f:
            f.write(f"include {CHANGELOG}\n")
    except FileNotFoundError:
        pass  # Ignore missing changelogs


def collect_package_data(base_path: Path) -> PackageData:
    """Generate package data for a setuptools.setup() call.

    This reflects the transformations done during copying in copy_stubs().
    """
    package_data: dict[str, list[str]] = {}
    for entry in base_path.iterdir():
        if is_ignored_distribution_file(entry):
            continue
        elif entry.is_file() and entry.suffix == ".pyi":
            pkg_name = entry.stem
            # Module -> package transformation is done while copying.
            stub_files = ["__init__.pyi"]
        elif entry.is_dir():
            pkg_name = entry.name
            stub_files = find_stub_files(str(entry))
        else:
            raise ValueError(f"Only stub files are allowed, not {entry.name!r}")
        package_data[pkg_name + SUFFIX] = [*stub_files, META]
    pkg_data = PackageData(base_path, package_data)
    for package in pkg_data.top_level_non_namespace_packages:
        pkg_data.add_file(package, "py.typed")
    return pkg_data


def is_ignored_distribution_file(entry: Path) -> bool:
    if entry.is_file():
        if entry.name == META:
            return True
        if entry.suffix in (".md", ".rst"):
            return True
        if (
            subprocess.run(
                ["git", "check-ignore", entry.name], cwd=str(entry.parent)
            ).returncode
            == 0
        ):
            return True
    elif entry.is_dir():
        if entry.name == TESTS_NAMESPACE:
            return True
    return False


def generate_setup_file(pkg_data: PackageData) -> str:
    """Auto-generate a setup.py file for given distribution using a template."""
    return SETUP_TEMPLATE.format(package_data=pkg_data.package_data)


def generate_pyproject_file(
    ts_data: TypeshedData,
    build_data: BuildData,
    pkg_data: PackageData,
    metadata: Metadata,
    version: str,
) -> str:
    """Auto-generate a setup.py file for given distribution using a template."""
    all_requirements = [
        str(req) for req in metadata.requires_typeshed + metadata.requires_external
    ]
    requires_python = (
        metadata.requires_python
        if metadata.requires_python is not None
        else f">={ts_data.oldest_supported_python}"
    )
    return PYPROJECT_TEMPLATE.format(
        distribution=build_data.distribution,
        stub_distribution=metadata.stub_distribution,
        long_description=generate_long_description(
            build_data.distribution, ts_data, metadata
        ),
        version=version,
        requires=all_requirements,
        packages=pkg_data.top_level_packages,
        requires_python=requires_python,
    )


def generate_long_description(
    distribution: str, ts_data: TypeshedData, metadata: Metadata
) -> str:
    commit = ts_data.read_current_commit()

    extra_description = metadata.extra_description.strip()
    parts: list[str] = []

    if metadata.upstream_repository is not None:
        formatted_distribution = f"[`{distribution}`]({metadata.upstream_repository})"
    else:
        formatted_distribution = f"`{distribution}`"

    parts.append(
        DESCRIPTION_INTRO_TEMPLATE.format(
            distribution=distribution,
            formatted_distribution=formatted_distribution,
            stub_distribution=metadata.stub_distribution,
            typeshed_version_spec=metadata.version_spec,
        )
    )
    if extra_description:
        parts.append(extra_description)
    if metadata.obsolete_since:
        parts.append(
            OBSOLETE_SINCE_TEXT_TEMPLATE.format(
                distribution=distribution,
                stub_distribution=metadata.stub_distribution,
                obsolete_since=metadata.obsolete_since,
            )
        )
    elif metadata.no_longer_updated:
        parts.append(
            NO_LONGER_UPDATED_TEMPLATE.format(
                stub_distribution=metadata.stub_distribution
            )
        )
    if metadata.partial:
        parts.append(PARTIAL_STUBS_DESCRIPTION)
    parts.append(
        DESCRIPTION_OUTRO_TEMPLATE.format(
            distribution=distribution,
            commit=commit,
            ts_data=ts_data,
        )
    )
    return "\n\n".join(parts)


def main(
    typeshed_dir: str, distribution: str, version: str, build_dir: str | None = None
) -> Path:
    """Generate a wheel for a third-party distribution in typeshed.

    Return the path to directory where wheel is created.

    Note: the caller should clean the temporary directory where wheel is
    created after uploading it.
    """
    ts_data = read_typeshed_data(Path(typeshed_dir))
    # TODO: Merge BuildData and PackageData and encapsulate MetaData.
    # See also https://github.com/typeshed-internal/stub_uploader/issues/123
    build_data = BuildData(typeshed_dir, distribution)
    metadata = read_metadata(typeshed_dir, distribution)
    pkg_data = collect_package_data(build_data.stub_dir)

    if build_dir:
        tmpdir = Path(build_dir)
    else:
        tmpdir = Path(tempfile.mkdtemp(prefix="stub-uploader-"))

    (tmpdir / "setup.py").write_text(generate_setup_file(pkg_data))
    (tmpdir / "pyproject.toml").write_text(
        generate_pyproject_file(ts_data, build_data, pkg_data, metadata, version)
    )
    copy_stubs(build_data.stub_dir, tmpdir)
    create_py_typed(metadata, pkg_data, tmpdir)
    (tmpdir / "README.md").write_text(
        generate_long_description(distribution, ts_data, metadata)
    )
    copy_license(ts_data, tmpdir)
    copy_changelog(distribution, tmpdir)

    print(f"\033[0;33mRunning '{sys.executable} -m build --no-isolation' in {tmpdir}")
    print()
    subprocess.run(["pip freeze | egrep 'build|twine'"], shell=True)
    print("\033[0m")
    subprocess.run(
        [sys.executable, "-m", "build", "--no-isolation"],
        cwd=tmpdir,
    )
    return Path(tmpdir) / "dist"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--build-dir", default=None, help="build directory")
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    parser.add_argument("distribution", help="Third-party distribution to build")
    parser.add_argument("version", help="New stub version")
    args = parser.parse_args()
    build_path = main(
        args.typeshed_dir, args.distribution, args.version, args.build_dir
    )
    print(f"Wheel is built in: {build_path}")



"""
Entry point for manual GitHub upload action.

This does following things:
* Finds all distributions with names that match the pattern provided
* Checks what is the current stub version increment for each package on PyPI
* Bumps the increment, builds and uploads the each new package to PyPI
* Verifies validity of stub dependencies, and updates known dependencies if needed
"""

import argparse
import os
import re

from stub_uploader.upload import upload


def main(typeshed_dir: str, pattern: str) -> None:
    """Force upload typeshed stub packages to PyPI."""
    compiled = re.compile(f"^{pattern}$")  # force exact matches
    matching = [
        d
        for d in os.listdir(os.path.join(typeshed_dir, "stubs"))
        if re.match(compiled, d)
    ]
    # TODO: Update changelog by providing the "commit" argument
    upload(typeshed_dir, matching)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    parser.add_argument(
        "pattern", help="Regular expression to select distributions for upload"
    )
    args = parser.parse_args()
    main(args.typeshed_dir, args.pattern)



"""
Find the latest typeshed increment for a stub package with given
(or current) version.

If the given version was never uploaded, this will return -1. See
https://github.com/python/typeshed/blob/main/README.md for details
on stub versioning.

This file also contains some helper functions related to querying
distribution information.
"""

from __future__ import annotations
import datetime

from typing import Any

import requests
from packaging.specifiers import Specifier, SpecifierSet
from packaging.version import Version
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from stub_uploader.const import TYPES_PREFIX
from stub_uploader.metadata import Metadata

URL_TEMPLATE = "https://pypi.org/pypi/{}/json"
RETRIES = 5
RETRY_ON = [429, 500, 502, 503, 504]
TIMEOUT = 3


def fetch_pypi_versions(distribution: str) -> list[Version]:
    assert distribution.startswith(TYPES_PREFIX)
    url = URL_TEMPLATE.format(distribution)
    retry_strategy = Retry(total=RETRIES, status_forcelist=RETRY_ON)
    with requests.Session() as session:
        session.mount("https://", HTTPAdapter(max_retries=retry_strategy))
        resp = session.get(url, timeout=TIMEOUT)
    if not resp.ok:
        if resp.status_code == 404:
            # Looks like this is first time this package is ever uploaded.
            return []
        raise ValueError("Error while retrieving version")
    releases: dict[str, Any] = resp.json()["releases"]
    return [Version(release) for release in releases.keys()]


def ensure_specificity(ver: list[int], specificity: int) -> None:
    ver.extend([0] * (specificity - len(ver)))


# `today` can be passed as an argument for testing purposes.
def compute_stub_version(
    version_spec: Specifier, published_versions: list[Version], today: datetime.date
) -> Version:
    # The most important postcondition is that the incremented version is greater than
    # all published versions. This ensures that users who don't pin get the most
    # up to date stub. If we ever maintain multiple versions for a stub, this will
    # need revisiting.
    max_published = max(published_versions, default=Version("0"))

    # Parse and massage the base version if necessary. Usually, the base version is
    # just the version_spec without any trailing `.*`. But if the version_spec is a
    # post version, we append the post version to the base version:
    #   1.2.post3 -> 1.2.3
    # This is necessary, since structured post versions (e.g. `1.2.post3.4`) are not
    # supported by PEP 440.
    version_base = Version(version_spec.version.removesuffix(".*"))
    if max_published.epoch > 0 or version_base.epoch > 0:
        raise NotImplementedError("Epochs in versions are not supported")
    elif version_base.is_prerelease or version_base.is_devrelease:
        raise NotImplementedError("Pre- and dev-releases in versions are not supported")
    elif version_base.is_postrelease:
        version_base = Version(f"{version_base.base_version}.{version_base.post}")

    # Usually, the new version should satisfy the version_spec (defined precisely by
    # the `compatible` specifier in `assert_compatibility()`). This allows users to
    # have expectations of what a stub package will contain based on the upstream
    # version they're targeting.
    is_compatible = True

    # We'll try to bump the fourth part of the release. So e.g. if our version_spec is
    # 1.1, we'll release 1.1.0.YYYYMMDD. But if our version_spec is 5.6.7.8, we'll
    # release 5.6.7.8.YYYYMMDD.
    specificity = len(version_base.release)
    stub_specificity = max(specificity + 1, 4)

    if version_base.release < max_published.release[:specificity]:
        # Our published versions have gone too far ahead the upstream version
        # So we can't guarantee our second property.
        # In practice, this will only happen if the specificity of version_spec is
        # changed or we change our versioning scheme.
        # For example, version_base=1.2, max_published=1.3.0.4, return 1.3.0.YYYYMMDD
        stub_specificity = max(stub_specificity, len(max_published.release))
        base_version_parts = list(max_published.release)

        # But can't keep versioning compatible with upstream...
        is_compatible = False
        # We also use the latest stub version as the base version, not the version_spec.
        base_version_changed = False

    elif version_base.release > max_published.release[:specificity]:
        # For example, version_base=1.2, max_published=1.1.0.4, return 1.2.0.YYYYMMDD
        base_version_parts = list(version_base.release)
        base_version_changed = True

    else:
        assert version_base.release == max_published.release[:specificity]
        # For example, version_base=1.1, max_published=1.1.0.YYYYMMDD (old),
        # return 1.1.0.YYYYMMDD (today)
        base_version_parts = list(max_published.release)
        base_version_changed = False

    ensure_specificity(base_version_parts, stub_specificity)
    unused_date = find_unused_date(max_published, base_version_changed, today)
    new_version_parts = [*base_version_parts[:-1], unused_date.strftime("%Y%m%d")]
    new_version = Version(".".join(map(str, new_version_parts)))
    assert_compatibility(
        version_base=version_base, new_version=new_version, is_compatible=is_compatible
    )
    assert (
        new_version > max_published
    ), f"new version {new_version} <= published version {max_published}"
    return new_version


# `today` can be passed as an argument for testing purposes.
def find_unused_date(
    max_published: Version, base_version_changed: bool, today: datetime.date
) -> datetime.date:
    # Is not necessarily a date in case the package predates the switch
    # to the current versioning scheme.
    last_date = max_published.release[-1]

    # If the base version changed, we should always be able to use the current date.
    if base_version_changed:
        return today

    # Make sure that the last published release is not unreasonable large.
    assert int(today.strftime("%Y%m%d")) + 14 > last_date

    new_date = today
    while int(new_date.strftime("%Y%m%d")) <= last_date:
        new_date += datetime.timedelta(days=1)
    return new_date


def assert_compatibility(
    *, version_base: Version, new_version: Version, is_compatible: bool
) -> None:
    compatible = SpecifierSet(f"=={version_base.base_version}.*")
    if is_compatible:
        assert new_version in compatible
    else:
        assert new_version not in compatible


def determine_stub_version(metadata: Metadata) -> str:
    today = datetime.datetime.now(datetime.UTC).date()
    published_stub_versions = fetch_pypi_versions(metadata.stub_distribution)
    version = compute_stub_version(
        metadata.version_spec, published_stub_versions, today
    )
    return str(version)



from __future__ import annotations

import subprocess

from stub_uploader import build_wheel
from stub_uploader.get_version import determine_stub_version
from stub_uploader.update_changelog import update_changelog
from stub_uploader.metadata import (
    read_metadata,
    recursive_verify,
    sort_by_dependency,
    uploaded_packages,
)


def upload(
    typeshed_dir: str,
    distributions: list[str],
    commit: str | None = None,
    *,
    dry_run: bool = False,
) -> None:
    to_upload = sort_by_dependency(typeshed_dir, distributions)
    print("Building and uploading stubs for:", ", ".join(distributions))
    print()
    for distribution in to_upload:
        upload_distribution(typeshed_dir, distribution, commit, dry_run=dry_run)


def upload_distribution(
    typeshed_dir: str, distribution: str, commit: str | None, *, dry_run: bool
) -> None:
    print(f"Building stubs for {distribution}... ", end="")
    metadata = read_metadata(typeshed_dir, distribution)
    recursive_verify(metadata, typeshed_dir)

    version = determine_stub_version(metadata)

    if commit:
        update_changelog(typeshed_dir, commit, distribution, version, dry_run=dry_run)

    temp_dir = build_wheel.main(typeshed_dir, distribution, version)
    print(f"ok, version {version}")

    print(f"Uploading stubs for {distribution}... ", end="")
    if dry_run or not metadata.upload:
        print("skipped")
    else:
        subprocess.run(["twine", "upload", str(temp_dir / "*")], check=True)
        uploaded_packages.add(distribution)
        print("ok")
    print()



"""Unit tests for simple helpers should go here."""

import datetime
import os
import tempfile
from io import StringIO
from pathlib import Path
from typing import Any

import pytest
from packaging.specifiers import Specifier
from packaging.version import Version

from stub_uploader.build_wheel import collect_package_data
from stub_uploader.get_version import compute_stub_version, ensure_specificity
from stub_uploader.metadata import Metadata, _UploadedPackages, strip_types_prefix
from stub_uploader.ts_data import TypeshedData, parse_requirements, read_typeshed_data
from stub_uploader.update_changelog import process_git_log


def test_strip_types_prefix() -> None:
    assert strip_types_prefix("types-foo") == "foo"
    with pytest.raises(ValueError):
        strip_types_prefix("bad")


def test_ensure_specificity() -> None:
    ver = [1]
    ensure_specificity(ver, 3)
    assert ver == [1, 0, 0]

    ver = [1, 2, 3]
    ensure_specificity(ver, 3)
    assert ver == [1, 2, 3]

    ver = [1, 2, 3, 4, 5]
    ensure_specificity(ver, 3)
    assert ver == [1, 2, 3, 4, 5]


TODAY = datetime.date(2024, 9, 12)
TODAY_V = TODAY.strftime("%Y%m%d")
TOMORROW_V = "20240913"
IN_TWO_DAYS_V = "20240914"


def _stub_ver(ver: str, published: list[str]) -> str:
    published_vers = [Version(v) for v in published]
    if ver[0].isdigit():
        ver = f"=={ver}"
    spec = Specifier(ver)
    return str(compute_stub_version(spec, published_vers, TODAY))


def test_compute_stub_version() -> None:
    # never before published version
    empty_list: list[str] = []
    assert _stub_ver("1", empty_list) == f"1.0.0.{TODAY_V}"
    assert _stub_ver("1.2", empty_list) == f"1.2.0.{TODAY_V}"
    assert _stub_ver("1.2.post3", empty_list) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("~=1.2", empty_list) == f"1.2.0.{TODAY_V}"
    assert _stub_ver("~=1.2.3", empty_list) == f"1.2.3.{TODAY_V}"

    # published greater than version spec
    assert _stub_ver("1.1", ["1.2"]) == f"1.2.0.{TODAY_V}"
    assert _stub_ver("1.1", ["1.2.3"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("1.2", ["1.3.0.4"]) == f"1.3.0.{TODAY_V}"
    assert _stub_ver("1.1", ["1.2.3.4.5"]) == f"1.2.3.4.{TODAY_V}"
    assert _stub_ver("1.4.40", ["1.4.50"]) == f"1.4.50.{TODAY_V}"
    assert _stub_ver("1.4.0.40", ["1.4.0.50"]) == f"1.4.0.50.{TODAY_V}"
    assert _stub_ver("1.2.post3", ["1.2.3.4"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("~=1.2", ["1.3.0.4"]) == f"1.3.0.{TODAY_V}"
    assert _stub_ver("~=1.2.3", ["1.2.4.40"]) == f"1.2.4.{TODAY_V}"

    # published less than version spec
    assert _stub_ver("1.2", ["1.1.0.4"]) == f"1.2.0.{TODAY_V}"
    assert _stub_ver("1", ["0.9"]) == f"1.0.0.{TODAY_V}"
    assert _stub_ver("1.1", ["0.9"]) == f"1.1.0.{TODAY_V}"
    assert _stub_ver("1.2.3", ["1.1.0.70"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("1.2.3", [f"1.1.0.{TODAY_V}"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("1.2.3", [f"1.1.0.{TOMORROW_V}"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("1.2.3.4", ["1.1.0.17"]) == f"1.2.3.4.{TODAY_V}"
    assert _stub_ver("1.2.post3", ["1.1.0.17"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("1.2.3", ["1.1.0.21000101"]) == f"1.2.3.{TODAY_V}"
    assert _stub_ver("~=1.2", ["1.1.0.40"]) == f"1.2.0.{TODAY_V}"
    assert _stub_ver("~=1.2.3", ["1.2.2.40"]) == f"1.2.3.{TODAY_V}"

    # published equals version spec
    assert _stub_ver("1.1", ["1.1"]) == f"1.1.0.{TODAY_V}"
    assert _stub_ver("1.1", ["1.1.0.19991204"]) == f"1.1.0.{TODAY_V}"
    assert _stub_ver("1.1", ["1.1.3.19991204"]) == f"1.1.3.{TODAY_V}"
    assert _stub_ver("1.2.3.4", ["1.2.3.4.19991204"]) == f"1.2.3.4.{TODAY_V}"
    assert _stub_ver("1.2.3.4.5", ["1.2.3.4.5"]) == f"1.2.3.4.5.{TODAY_V}"
    assert _stub_ver("~=1.2.3", ["1.2.3.40"]) == f"1.2.3.{TODAY_V}"

    # test with multiple published versions
    assert (
        _stub_ver("1.2", ["1.1.0.7", "1.2.0.7", "1.3.0.19991204"]) == f"1.3.0.{TODAY_V}"
    )

    # today's package was already uploaded
    assert (
        _stub_ver("1.2.*", ["1.2.0.1", f"1.2.0.{TODAY_V}", f"1.2.0.{TOMORROW_V}"])
        == f"1.2.0.{IN_TWO_DAYS_V}"
    )
    assert (
        _stub_ver("1.2.*", [f"1.1.0.{TODAY_V}", f"1.1.0.{TOMORROW_V}"])
        == f"1.2.0.{TODAY_V}"
    )
    assert (
        _stub_ver("~=1.2.3", [f"1.2.3.{TODAY_V}", f"1.2.3.{TOMORROW_V}"])
        == f"1.2.3.{IN_TWO_DAYS_V}"
    )
    assert _stub_ver("~=1.2.3", [f"1.4.0.{TODAY_V}"]) == f"1.4.0.{TOMORROW_V}"


def test_collect_package_data() -> None:
    stubs = Path("data") / "test_typeshed" / "stubs"
    pkg_data = collect_package_data(stubs / "singlefilepkg")
    assert pkg_data.top_level_packages == ["singlefilepkg-stubs"]
    assert pkg_data.package_data == (
        {"singlefilepkg-stubs": ["__init__.pyi", "METADATA.toml", "py.typed"]}
    )

    pkg_data = collect_package_data(stubs / "multifilepkg")
    assert pkg_data.top_level_packages == ["multifilepkg-stubs"]
    assert pkg_data.package_data == (
        {
            "multifilepkg-stubs": [
                "__init__.pyi",
                "a.pyi",
                "b.pyi",
                os.path.join("c", "__init__.pyi"),
                os.path.join("c", "d.pyi"),
                os.path.join("c", "e.pyi"),
                "METADATA.toml",
                "py.typed",
            ]
        }
    )

    pkg_data = collect_package_data(stubs / "nspkg")
    assert pkg_data.top_level_packages == ["nspkg-stubs"]
    assert pkg_data.package_data == (
        {
            "nspkg-stubs": [
                os.path.join("innerpkg", "__init__.pyi"),
                "METADATA.toml",
                os.path.join("innerpkg", "py.typed"),
            ]
        }
    )


def test_collect_package_data_bogusfile() -> None:
    stubs = Path("data") / "test_typeshed" / "stubs"
    with pytest.raises(
        ValueError, match="Only stub files are allowed, not 'bogusfile.txt'"
    ):
        collect_package_data(stubs / "bogusfiles")

    # Make sure gitignored files aren't collected, nor do they crash function
    with open(os.path.join(stubs, "singlefilepkg", ".METADATA.toml.swp"), "w"):
        pass
    pkg_data = collect_package_data(stubs / "singlefilepkg")
    assert set(pkg_data.package_data["singlefilepkg-stubs"]) == {
        "__init__.pyi",
        "py.typed",
        "METADATA.toml",
    }

    with open(
        os.path.join(stubs, "multifilepkg", "multifilepkg", ".METADATA.toml.swp"), "w"
    ):
        pass
    pkg_data = collect_package_data(stubs / "multifilepkg")
    assert set(pkg_data.package_data["multifilepkg-stubs"]) == {
        "__init__.pyi",
        "a.pyi",
        "b.pyi",
        "c/__init__.pyi",
        "c/d.pyi",
        "c/e.pyi",
        "py.typed",
        "METADATA.toml",
    }


def test_uploaded_packages() -> None:
    with tempfile.TemporaryDirectory() as td:
        file_path = os.path.join(td, "uploaded_packages.txt")
        with open(file_path, "w") as f:
            f.write("types-SqLaLcHeMy\n")

        up = _UploadedPackages(file_path)
        assert up.read() == {"types-sqlalchemy"}

        up.add("SQLAlchemy")
        assert up.read() == {"types-sqlalchemy"}

        up.add("six")
        assert up.read() == {"types-sqlalchemy", "types-six"}

        with open(file_path) as f:
            assert f.read() == "types-SqLaLcHeMy\ntypes-six"


def _build_pyproject(oldest_supported_python: str | None = "3.10") -> str:
    pp = "[tool.typeshed]\n"
    if oldest_supported_python is not None:
        pp += f'oldest_supported_python = "{oldest_supported_python}"\n'
    return pp


_REQUIREMENTS_TXT = """# This is a comment

pkg1==1.2
pkg2==2.3.4  # This is a comment
pkg3==2023.4.13; python_version >= "3.6"
multispec==3.4.5,>3.4.5
range>=1.2.3
no_version
"""


def _build_requirements(
    *,
    mypy: str | None = "1.11.1",
    pyright: str | None = "1.1.381",
) -> str:
    req = _REQUIREMENTS_TXT
    if mypy is not None:
        req += f"\nmypy=={mypy}"
    if pyright is not None:
        req += f"\npyright=={pyright}"
    return req


def _test_typeshed_data(pyproject: str, requirements: str) -> TypeshedData:
    with tempfile.TemporaryDirectory() as td:
        temp_path = Path(td)
        (temp_path / "pyproject.toml").write_text(pyproject)
        (temp_path / "requirements-tests.txt").write_text(requirements)
        return read_typeshed_data(temp_path)


def test_read_typeshed_data__success() -> None:
    pp = _build_pyproject(oldest_supported_python="3.10")
    req = _build_requirements(mypy="1.11.1")
    data = _test_typeshed_data(pp, req)
    assert data.oldest_supported_python == "3.10"
    assert data.mypy_version == Version("1.11.1")


def test_read_typeshed_data__oldest_python_missing() -> None:
    pp = _build_pyproject(oldest_supported_python=None)
    req = _build_requirements(mypy=None)
    with pytest.raises(KeyError, match="oldest_supported_python"):
        _test_typeshed_data(pp, req)


def test_read_typeshed_data__oldest_python_invalid() -> None:
    pp = _build_pyproject(oldest_supported_python="3.12invalid")
    req = _build_requirements(mypy=None)
    with pytest.raises(ValueError, match="oldest_supported_python"):
        _test_typeshed_data(pp, req)


def test_read_typeshed_data__type_checkers_missing() -> None:
    pp = _build_pyproject()
    req = _build_requirements(mypy=None)
    with pytest.raises(KeyError, match="mypy"):
        _test_typeshed_data(pp, req)


def test_read_typeshed_data__invalid_type_checker_version() -> None:
    pp = _build_pyproject()
    req = _build_requirements(mypy="1.2.2invalid")
    with pytest.raises(ValueError, match="mypy"):
        _test_typeshed_data(pp, req)


@pytest.mark.parametrize(
    "name,version",
    [
        ("pkg1", "1.2"),
        ("pkg2", "2.3.4"),
        ("pkg3", "2023.4.13"),
    ],
)
def test_parse_requirements__parsed_packages(name: str, version: str) -> None:
    requirements = parse_requirements(StringIO(_REQUIREMENTS_TXT))
    assert name in requirements, f"package {name} not found"
    assert requirements[name] == version, f"package {name} has wrong version"


@pytest.mark.parametrize(
    "name",
    [
        "multispec",
        "range",
        "no_version",
    ],
)
def test_parse_requirements__skipped_packages(name: str) -> None:
    requirements = parse_requirements(StringIO(_REQUIREMENTS_TXT))
    assert name not in requirements, f"package {name} was not skipped"


@pytest.mark.parametrize(
    "data,expected",
    [
        ({}, None),
        ({"upstream_repository": 12345}, None),
        ({"upstream_repository": "https://[].foo.com"}, None),
        (
            {"upstream_repository": "https://github.com/psf/requests"},
            "https://github.com/psf/requests",
        ),
    ],
)
def test_upstream_repo_validation(data: dict[str, Any], expected: str | None) -> None:
    m = Metadata("foo", data)
    assert m.upstream_repository == expected
    assert type(m.upstream_repository) is type(expected)


def test_process_git_log() -> None:
    git_log = """
commit 126768408a69b7a3a09b7d3992970b289f92937e

    Replace `Incomplete | None = None` in third party stubs (#14063)

commit 337fd828e819988af2d3600283d8068bbbab7f50

    Bump setuptools to 80.7.* (#14069)
    
    Body text #12345 python/typeshed#12345 https://example.com/#123
"""
    expected_entry = f"""\
## 80.7.0.{TODAY_V} ({TODAY:%Y-%m-%d})

Replace `Incomplete | None = None` in third party stubs ([#14063](https://github.com/python/typeshed/pull/14063))

Bump setuptools to 80.7.* ([#14069](https://github.com/python/typeshed/pull/14069))

Body text #12345 python/typeshed#12345 https://example.com/#123

"""
    actual_entry = process_git_log(git_log, f"80.7.0.{TODAY_V}", TODAY)
    assert actual_entry == expected_entry



"""
Entry point for scheduled GitHub auto-upload action.

This does following things:
* Reads the list of stub packages modified since last commit in typeshed
* Checks what is the current stub version increment for each package on PyPI
* Bumps the increment, builds and uploads (unless run with --dry-run) each
  new package to PyPI
* Verifies validity of stub dependencies, and updates known dependencies if needed
"""

import argparse
import os
from stub_uploader import get_changed
from stub_uploader.upload import upload


def main(typeshed_dir: str, commit: str, dry_run: bool = False) -> None:
    """Upload stub typeshed packages modified since commit."""
    changed = get_changed.main(typeshed_dir, commit)
    # Ignore those distributions that were completely deleted.
    current = set(os.listdir(os.path.join(typeshed_dir, "stubs")))
    changed = [d for d in changed if d in current]
    upload(typeshed_dir, changed, commit, dry_run=dry_run)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    parser.add_argument(
        "previous_commit", help="Previous typeshed commit for which we performed upload"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Should we perform a dry run (don't actually upload)",
    )
    args = parser.parse_args()
    main(args.typeshed_dir, args.previous_commit, args.dry_run)



"""
Integration tests for build scripts. These should not change
anything on PyPI, but can make PyPI queries and may expect
a typeshed checkout side by side.
"""

import os
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from stub_uploader import build_wheel, get_version
from stub_uploader.const import THIRD_PARTY_NAMESPACE
from stub_uploader.metadata import (
    InvalidRequires,
    Metadata,
    read_metadata,
    recursive_verify,
    sort_by_dependency,
    strip_types_prefix,
    verify_external_req,
    verify_requires_python,
    verify_typeshed_req,
)
from stub_uploader.ts_data import read_typeshed_data

TYPESHED = "../typeshed"
THIRD_PARTY_PATH = Path(TYPESHED) / THIRD_PARTY_NAMESPACE


def test_fetch_pypi_versions() -> None:
    """Check that we can query PyPI for package increments."""
    assert Version("1.16.0") in get_version.fetch_pypi_versions("types-six")
    assert Version("1.5.4") in get_version.fetch_pypi_versions("types-typed-ast")
    assert not get_version.fetch_pypi_versions("types-nonexistent-distribution")


@pytest.mark.parametrize("distribution", os.listdir(THIRD_PARTY_PATH))
def test_build_wheel(distribution: str) -> None:
    """Check that we can build wheels for all distributions."""
    with TemporaryDirectory(prefix="stub-uploader-") as tmp_dir:
        build_path = build_wheel.main(
            TYPESHED, distribution, version="1.1.1", build_dir=tmp_dir
        )
        assert build_path.name == "dist"
        assert list(build_path.iterdir())  # check it is not empty


@pytest.mark.parametrize(
    "distribution,file_list",
    [
        ("flake8", [Path("flake8-stubs") / "checker.pyi"]),
        ("protobuf", [Path("google-stubs") / "protobuf" / "__init__.pyi"]),
    ],
)
def test_build_wheel_files(distribution: str, file_list: list[Path]) -> None:
    """Assert that a wheel contains certain files."""
    with TemporaryDirectory(prefix="stub-uploader-") as tmp_dir:
        build_path = build_wheel.main(
            TYPESHED, distribution, version="1.1.1", build_dir=tmp_dir
        )
        wheels = list(build_path.glob("*.whl"))
        assert len(wheels) == 1
        with ZipFile(wheels[0], "r") as zip_file:
            zip_file.extractall(build_path)
        for file_path in file_list:
            assert build_path.joinpath(file_path).exists()


@pytest.mark.parametrize("distribution", os.listdir(THIRD_PARTY_PATH))
def test_version_increment(distribution: str) -> None:
    get_version.determine_stub_version(read_metadata(TYPESHED, distribution))


def test_unvalidated_properties() -> None:
    m = Metadata("fake", {"version": "0.1", "requires": ["numpy", "types-six>=0.1"]})
    assert [r.name for r in m._unvalidated_requires] == ["numpy", "types-six"]
    assert [r.name for r in m._unvalidated_requires_external] == ["numpy"]
    assert [r.name for r in m._unvalidated_requires_typeshed] == ["types-six"]


def test_verify_typeshed_req() -> None:
    # Check that some known dependencies verify as valid.
    verify_typeshed_req(Requirement("types-six"))
    verify_typeshed_req(Requirement("types-six==0.1.1"))
    verify_typeshed_req(Requirement("types-typed-ast"))
    verify_typeshed_req(Requirement("types-typed-ast>=3.7"))

    with pytest.raises(InvalidRequires, match="to start with types-"):
        verify_typeshed_req(Requirement("unsupported"))

    with pytest.raises(InvalidRequires, match="to be uploaded from stub_uploader"):
        verify_typeshed_req(Requirement("types-unknown-xxx"))

    m = Metadata("mypy", {"version": "0.1", "requires": ["types-unknown-xxx"]})
    assert m.requires_typeshed == []


def test_verify_external_req() -> None:
    # Check that some known dependencies verify as valid.
    verify_external_req(
        Requirement("typing-extensions"), "mypy", _unsafe_ignore_allowlist=True
    )
    verify_external_req(
        Requirement("mypy-extensions"), "mypy", _unsafe_ignore_allowlist=True
    )
    # Check that types-foo can depend on foo
    verify_external_req(Requirement("setuptools"), "setuptools")

    with pytest.raises(
        InvalidRequires, match="to be present in the stub_uploader allowlist"
    ):
        verify_external_req(Requirement("typing-extensions"), "mypy")

    m = Metadata("pandas", {"version": "0.1", "requires": ["numpy"]})
    assert [r.name for r in m.requires_external] == ["numpy"]

    with pytest.raises(InvalidRequires, match="to be listed in mypy's requires_dist"):
        verify_external_req(Requirement("numpy"), "mypy")

    with pytest.raises(InvalidRequires, match="to not be uploaded from stub_uploader"):
        verify_external_req(Requirement("types-typed-ast"), "mypy")

    with pytest.raises(InvalidRequires, match="to not start with types-"):
        verify_external_req(Requirement("types-unknown-xxx"), "mypy")

    m = Metadata("mypy", {"version": "0.1", "requires": ["numpy"]})
    with pytest.raises(InvalidRequires, match="to be listed in mypy's requires_dist"):
        m.requires_external
    with pytest.raises(InvalidRequires, match="to be listed in mypy's requires_dist"):
        recursive_verify(m, TYPESHED)

    # TODO: change tests once METADATA.toml specifies whether a dist is on PyPI
    m = Metadata("gdb", {"version": "0.1", "requires": []})
    assert m.requires_external == []

    m = Metadata("gdb", {"version": "0.1", "requires": ["cryptography"]})
    with pytest.raises(InvalidRequires, match="no upstream distribution on PyPI"):
        m.requires_external

    # Check differing runtime and stub dependencies
    verify_external_req(Requirement("pandas-stubs"), "geopandas")
    with pytest.raises(
        InvalidRequires,
        match=(
            r"Expected dependency pandas to be present in the stub_uploader allowlist"
            r"\. Did you mean pandas-stubs\?"
        ),
    ):
        verify_external_req(Requirement("pandas"), "geopandas")


def test_dependency_order() -> None:
    """Test sort_by_dependency correctly sorts all packages by dependency."""
    distributions = os.listdir(os.path.join(TYPESHED, "stubs"))
    to_upload = list(sort_by_dependency(TYPESHED, distributions))
    assert len(set(to_upload)) == len(to_upload)
    for distribution in distributions:
        for req in read_metadata(TYPESHED, distribution).requires_typeshed:
            assert to_upload.index(strip_types_prefix(req.name)) < to_upload.index(
                distribution
            )


def test_recursive_verify_single() -> None:
    m = read_metadata(TYPESHED, "six")
    assert recursive_verify(m, TYPESHED) == {"types-six"}

    m = read_metadata(TYPESHED, "requests-oauthlib")
    assert recursive_verify(m, TYPESHED) == {
        "types-requests-oauthlib",
        "types-requests",
        "types-oauthlib",
    }


def test_dependency_order_single() -> None:
    assert list(sort_by_dependency(TYPESHED, ["requests-oauthlib"])) == [
        "requests-oauthlib"
    ]
    assert list(sort_by_dependency(TYPESHED, ["requests-oauthlib", "oauthlib"])) == [
        "oauthlib",
        "requests-oauthlib",
    ]


@pytest.mark.parametrize("distribution", os.listdir(THIRD_PARTY_PATH))
def test_recursive_verify(distribution: str) -> None:
    recursive_verify(read_metadata(TYPESHED, distribution), TYPESHED)


def test_read_typeshed_data() -> None:
    read_typeshed_data(Path(TYPESHED))


def test_verify_requires_python() -> None:
    verify_requires_python(">=3.10")

    with pytest.raises(InvalidRequires, match="Invalid requires_python specifier"):
        verify_requires_python(">=fake")

    with pytest.raises(
        InvalidRequires, match="Expected requires_python to be a '>=' specifier"
    ):
        verify_requires_python("==3.10")


@pytest.mark.parametrize(
    "distribution,expected_packages",
    [
        ("fpdf2", ["fpdf-stubs"]),
        ("pytz", ["pytz-stubs"]),
        ("protobuf", ["google-stubs"]),
        ("google-cloud-ndb", ["google-stubs"]),
    ],
)
def test_pkg_data_top_level_packages(
    distribution: str, expected_packages: list[str]
) -> None:
    pkg_data = build_wheel.collect_package_data(THIRD_PARTY_PATH / distribution)
    assert pkg_data.top_level_packages == expected_packages


@pytest.mark.parametrize(
    "distribution,expected_packages",
    [
        ("fpdf2", ["fpdf-stubs"]),
        ("pytz", ["pytz-stubs"]),
        ("protobuf", ["google-stubs.protobuf"]),
        ("google-cloud-ndb", ["google-stubs.cloud.ndb"]),
    ],
)
def test_pkg_data_non_namespace_packages(
    distribution: str, expected_packages: list[str]
) -> None:
    pkg_data = build_wheel.collect_package_data(THIRD_PARTY_PATH / distribution)
    assert pkg_data.top_level_non_namespace_packages == expected_packages



#!/usr/bin/env python3

"""
This script updates the changelog for a single module.
"""

import argparse
import datetime
import os
import re
import subprocess
from pathlib import Path

from stub_uploader.const import CHANGELOG_PATH

THIRD_PARTY_NAMESPACE = "stubs"


def main() -> None:
    args = parse_args()
    update_changelog(
        args.typeshed_dir,
        args.previous_commit,
        args.distribution,
        args.version,
        dry_run=args.dry_run,
    )


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    parser.add_argument(
        "previous_commit", help="Previous typeshed commit for which we performed upload"
    )
    parser.add_argument("distribution", help="Distribution to update the changelog for")
    parser.add_argument("version", help="New version to add to the changelog")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Should we perform a dry run (don't actually update)",
    )
    return parser.parse_args()


def update_changelog(
    typeshed_dir: str,
    previous_commit: str,
    distribution: str,
    version: str,
    *,
    dry_run: bool = False,
) -> None:
    dist_path = Path(THIRD_PARTY_NAMESPACE, distribution)
    result = subprocess.run(
        ["git", "log", f"{previous_commit}..HEAD", "--", dist_path],
        cwd=typeshed_dir,
        stdout=subprocess.PIPE,
        check=True,
    )
    log = result.stdout.decode("utf-8")
    if not log:
        print(f"{distribution}: Changelog unchanged")
        return
    today = datetime.date.today()
    new_entry = process_git_log(log, version, today)

    changelog = os.path.join(CHANGELOG_PATH, f"{distribution}.md")
    try:
        with open(changelog) as f:
            old_entries = f.read()
    except FileNotFoundError:
        old_entries = ""

    if dry_run:
        if old_entries:
            print(
                f"Would add {len(new_entry.splitlines())} lines to existing {changelog}"
            )
        else:
            print(f"Would add {len(new_entry.splitlines())} lines to new {changelog}")
        return

    with open(changelog, "w") as f:
        f.write(new_entry + old_entries)


def process_git_log(log: str, version: str, today: datetime.date) -> str:
    entry = f"## {version} ({today:%Y-%m-%d})\n"
    for line in log.splitlines():
        if line.strip() == "":
            entry += "\n"
        elif line.startswith("    "):
            # Proper git log lines start with four spaces.
            entry += f"{line.rstrip()[4:]}\n"
        else:
            pass  # Ignore header entries.
    entry += "\n"
    # Create hyperlinks for PR numbers.
    entry = re.sub(
        r"\(#(\d+)\)", r"([#\1](https://github.com/python/typeshed/pull/\1))", entry
    )
    # Strip multiple empty lines.
    return re.sub("\n\n+", "\n\n", entry)


if __name__ == "__main__":
    main()



import pathlib

THIRD_PARTY_NAMESPACE = "stubs"
TESTS_NAMESPACE = "@tests"
META = "METADATA.toml"
TYPES_PREFIX = "types-"

_ROOT = pathlib.Path(__file__).parent.parent
CHANGELOG_PATH = (_ROOT / "data" / "changelogs").resolve()
UPLOADED_PATH = str((_ROOT / "data" / "uploaded_packages.txt").resolve())



import os

from setuptools import setup

increment = os.environ["PACKAGE_VERSION_INCREMENT"]
name = "types-tsbot-empty"
description = "Empty typing stubs for testing typeshed/PyPI auto-upload"

setup(
    name=name,
    version=f"0.1.{increment}",
    description=description,
    long_description=description,
    url="https://github.com/typeshed-internal/stub_uploader",
    install_requires=[],
    packages=[],
    classifiers=[
        "Typing :: Stubs Only",
    ],
)



"""
Information about typeshed.
"""

from __future__ import annotations

import argparse
from collections.abc import Iterable
from dataclasses import dataclass, fields
from pathlib import Path
import re
import subprocess
import tomllib

from packaging.requirements import Requirement
from packaging.version import Version

REQUIREMENTS = "requirements-tests.txt"
PYPROJECT = "pyproject.toml"


@dataclass
class TypeshedData:
    typeshed_path: Path
    mypy_version: Version
    pyright_version: Version
    oldest_supported_python: str

    def read_current_commit(self) -> str:
        return subprocess.run(
            ["git", "rev-parse", "HEAD"],
            capture_output=True,
            text=True,
            cwd=self.typeshed_path,
        ).stdout.strip()


def read_typeshed_data(typeshed_path: Path) -> TypeshedData:
    with (typeshed_path / PYPROJECT).open("rb") as f:
        pyproject = tomllib.load(f)
    with (typeshed_path / REQUIREMENTS).open() as f:
        requirements = parse_requirements(f)
    typeshed_table = pyproject["tool"]["typeshed"]
    if not re.match(r"^\d+\.\d+$", typeshed_table["oldest_supported_python"]):
        raise ValueError("Invalid oldest_supported_python in pyproject.toml")
    return TypeshedData(
        typeshed_path=typeshed_path,
        mypy_version=Version(requirements["mypy"]),
        pyright_version=Version(requirements["pyright"]),
        oldest_supported_python=typeshed_table["oldest_supported_python"],
    )


def parse_requirements(stream: Iterable[str]) -> dict[str, str]:
    """Parse all exact requirements from a requirements.txt file.

    Only requirements that are pinned to a specific version are returned.
    """
    requirements = {}
    for line in stream:
        line = line.strip()
        line = line.split("#")[0]  # strip comments
        if not line.strip():  # skip empty lines
            continue
        req = Requirement(line.strip())
        if len(req.specifier) != 1:
            continue
        spec = next(iter(req.specifier))
        if spec.operator != "==":
            continue
        requirements[req.name] = spec.version
    return requirements


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("typeshed_dir", help="Path to typeshed checkout directory")
    args = parser.parse_args()
    data = read_typeshed_data(Path(args.typeshed_dir))
    for field in fields(data):
        print(f"{field.name}: {getattr(data, field.name)}")
